@@ -261,14 +261,45 @@ class Scalar(Variable):
 
     def to_json_schema(self):
         rv = super(Scalar, self).to_json_schema()
-        rv.update({
-            'anyOf':  [
-                {'type': 'string'},
-                {'type': 'number'},
-                {'type': 'boolean'},
-                {'type': 'null'},
-            ],
-        })
+        rv['anyOf'] = [
+            {'type': 'boolean'},
+            {'type': 'null'},
+            {'type': 'number'},
+            {'type': 'string'},
+        ]
+        return rv
+
+
+class String(Scalar):
+    """A string."""
+    def __repr__(self):
+        return '<string>'
+
+    def to_json_schema(self):
+        rv = Variable.to_json_schema(self)
+        rv['type'] = 'string'
+        return rv
+
+
+class Number(Scalar):
+    """A number."""
+    def __repr__(self):
+        return '<number>'
+
+    def to_json_schema(self):
+        rv = Variable.to_json_schema(self)
+        rv['type'] = 'number'
+        return rv
+
+
+class Boolean(Scalar):
+    """A number."""
+    def __repr__(self):
+        return '<boolean>'
+
+    def to_json_schema(self):
+        rv = Variable.to_json_schema(self)
+        rv['type'] = 'boolean'
         return rv
 
 
@@ -279,14 +310,12 @@ class Unknown(Variable):
 
     def to_json_schema(self):
         rv = super(Unknown, self).to_json_schema()
-        rv.update({
-            'anyOf':  [
-                {'type': 'object'},
-                {'type': 'array'},
-                {'type': 'string'},
-                {'type': 'number'},
-                {'type': 'boolean'},
-                {'type': 'null'},
-            ],
-        })
+        rv['anyOf'] = [
+            {'type': 'object'},
+            {'type': 'array'},
+            {'type': 'string'},
+            {'type': 'number'},
+            {'type': 'boolean'},
+            {'type': 'null'},
+        ]
         return rv
